/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "TableModel.h"

/**
 * \brief Constructs an empty model.
 *
 * \param parent Pointer on the parent QWidget.
 */
TableModel::TableModel ( QWidget *parent ) :
    QAbstractTableModel ( parent )
{
    oldRows_      = 0;
    oldColumns_   = 0;
    blockList_    = 0;
    viewColumns_  = 0;
    parentWidget_ = parent;
}

/**
 * \brief Adds a block to the model.
 *
 * Blocks already contained in the model, with the same name as the new blocks
 * will always be replaced. If the warnings argument is true, a newWarning
 * signal is emitted in the case of a replacement.
 * Emits blocksChanged.
 *
 * \param name Name of the block.
 * \param start Lower boundary of the block (First value of codepoint).
 * \param end Upper boundary of the block (First value of codepoint).
 * \param warnings Enables or disable the warning.
 * \see showBlocks()
 */
void TableModel::addBlock ( QString name, uint start, uint end, bool warnings )
{
    // Check if the model has already a block with this name and remove it
    if ( blocks_.keys ().contains ( name )
            || blockItems_.keys ().contains ( name ) )
    {
        blocks_.remove     ( name );
        blockItems_.remove ( name );

        if ( warnings )
        {
            emit newWarning ( tr ( "Replaced block \"%1\"." ).arg ( name ) );
        }
    }

    // store the name and the boundaries
    blocks_.insertMulti ( name, end );
    blocks_.insertMulti ( name, start );

    uint codepoint;
    // Find all letters that belong into the block
    // insertMulti puts new items in front
    for ( int i = items_.size () - 1; i >= 0; --i )
    {
        if ( items_.at ( i ).getCodepoint ().isEmpty () )
        {
            continue;
        }
        codepoint = items_.at ( i ).getCodepoint ().at ( 0 );
        if ( codepoint >= start && codepoint <= end )
        {
            blockItems_.insertMulti ( name, &items_[ i ] );
            items_[ i ].setBlock ( name );
        }
    }
    emit blocksChanged ( blocks_ );
}

/**
 * \brief Adds a letter to the model.
 *
 * If the overwrite argument is true, the model checks if it
 * already contains a letter with the same codepoint, but different collation
 * or comment. Duplicate letters will then be replaced and a newWarning signal
 * emitted.
 * If the overwrite argument is false, the new letter will be appended to the
 * model regardless of duplicates.
 * The letter will be added to a block if the codepoint falls into a blocks
 * boundary.
 *
 * \param letter The letter to add.
 * \param overwrite If true, a letter with the same codepoint will be
 *        replaced.
 *
 */
void TableModel::addLetter ( Letter letter, bool overwrite )
{
    Letter* oldLetter;
    // check if the model already contains the letter
    if ( overwrite && indexMap.contains ( letter ) )
    {
        int idx = indexMap.value ( letter );
        oldLetter = &items_[ idx ];
        // model contains a letter with same codepoint otherwise different
        if ( !oldLetter->equals ( letter ) )
        {
            // emit a warning for the warnigs dialog
            emit newWarning ( tr ( " Replace \"%1\" from table \"%2\""
                                   " with \"%3\" from table \"%4\"." )
                                      .arg ( oldLetter->getCodepointAsHex () )
                                      .arg ( oldLetter->getVersionInfo    () )
                                      .arg ( letter.getCodepointAsHex     () )
                                      .arg ( letter.getVersionInfo        () ) );

            // replace the letter and update the indexMap
            items_.replace ( idx, letter );
            indexMap.remove ( *oldLetter );
            indexMap.insert ( letter, idx );
        }
    } else // just append the letter
    {
        items_.append ( letter );
        indexMap.insert ( letter, items_.size () - 1 );
    }
    // put the new letter in its block
    updateBlocks ();
}

/**
 * \brief Adds a list of letters to the model.
 *
 * If the overwrite argument is true, the model checks for each letter, if it
 * already contains a letter with the same codepoint, but different collation
 * or comment. Duplicate letters will be replaced and a newWarning signal
 * emitted.
 * If the overwrite argument is false, the new letters will be appended to the
 * model regardless of duplicates.
 * All letters will be added to their blocks if their codepoints fall into a
 * boundary.
 *
 * \param list The list of letters to add.
 * \param overwrite If true, letters with the same codepoint will be
 *        replaced.
 */
void TableModel::addLetterList ( QList<Letter> list, bool overwrite )
{
    QProgressDialog progress ( tr ( "Loading table..." ), tr ( "&Cancel" ), 0,
                               list.size (), parentWidget_ );
#ifndef Q_WS_MAC
    progress.setWindowModality ( Qt::ApplicationModal );
#endif

    Letter * newLetter;
    Letter * oldLetter;

    for ( int i = 0; i < list.size (); ++i )
    {
        progress.setValue ( i );
        QApplication::processEvents ();
        if ( progress.wasCanceled () )
            break;

        newLetter = &list[ i ];
        // check if the model already contains the letter and overwrite is true
        if ( overwrite && indexMap.contains ( *newLetter ) )
        {
            int idx = indexMap.value ( *newLetter );
            oldLetter = &items_[ idx ];
            // model contains a letter with same codepoint otherwise different
            if ( !oldLetter->equals ( *newLetter ) )
            {
                emit newWarning ( tr ( "Replace \"%1\" from table \"%2\" "
                                       "with \"%3\" from table \"%4\"." )
                                    .arg ( oldLetter->getCodepointAsHex () )
                                    .arg ( oldLetter->getVersionInfo    () )
                                    .arg ( newLetter->getCodepointAsHex () )
                                    .arg ( newLetter->getVersionInfo    () ) );
                // replace the letter and update the indexMap
                items_.replace ( idx, *newLetter );
                indexMap.remove ( *oldLetter );
                indexMap.insert ( *newLetter, idx );
            }
        } else // just append the letter
        {
            items_.append ( *newLetter );
            indexMap.insert ( *newLetter, items_.size () - 1 );
        }
    }

    progress.setValue ( list.size () );
    updateBlocks ();
}

/**
 * \brief Adds a list of letters to the view.
 *
 * Does not clear the view or check for duplicates.
 * Emits visibleChanged.
 *
 * \param list List of letters to show.
 */
void TableModel::addList ( QList<Letter*> list )
{
    visibleItems_.append ( list );
    recalcLayout ();
    emit visibleChanged ();
}

/**
 * \brief Adds a range of codepoints in the view.
 *
 * Only the first value of the codepoints are taken into account in the
 * range boundaries. Emits visibleChanged.
 *
 * \param start First/lowest codepoint.
 * \param end Last/highest codepoint.
 */
void TableModel::addRange ( uint start, uint end )
{
    for ( int i = 0; i < items_.size (); i++ )
    {
        if ( !items_.at ( i ).getCodepoint ().isEmpty () )
        {
            uint codepoint = items_.at ( i ).getCodepoint ().at ( 0 );
            if ( codepoint >= start && codepoint <= end )
                visibleItems_.append ( &items_[ i ] );
        }
    }
    recalcLayout ();
    emit visibleChanged ();
}

/**
 * \brief Adds letters within the range of two weights to the view.
 *
 * Collations consist of up to four levels and may be variable. The flags
 * argument defines which levels are taken into account. A Letter can have
 * multiple collations. The collationsIndices argument defines which collations
 * are taken into account. The view will not be cleared, there may be
 * duplicates. Emits visibleChanged.
 *
 * \param begin First/lowest collation.
 * \param end Last/highest collation.
 * \param flags Defines which levels define the range. Use values of the enum
 *              collationMode.
 * \param collationIndices Defines which collations define the range it letters
 *                         have multiple collations.
 */
void TableModel::addWeightRange ( QStringList begin, QStringList end,
                                  TableModel::CollationModes flags,
                                  QList<int> collationIndices )
{
    if ( begin.size () != end.size () )
    {
        qFatal ( "addWeightRange: The begin and end of"
                 " ranges have not the same size."      );
    }
    bool number = true;
    // Array for the begin and end values.
    uint weightbegin[ begin.size () ];
    uint weightend[ end.size () ];

    // If the flag COLLATION_STRING isn't set, the arrays will be filled in with begin and
    // end values.
    if ( TableModel::COLLATION_STRING
            != ( flags & TableModel::COLLATION_STRING ) )
    {
        for ( int i = 0; i < begin.size (); i++ )
        {
            weightbegin[ i ] = begin.at ( i ).toUInt ( &number, 16 );
            if ( !number )
                return;

            weightend[ i ] = end.at ( i ).toUInt ( &number, 16 );

            if ( !number )
                return;
        }
    }
    QList<Letter*> list;

    int cnt = 0;
    if ( TableModel::COLLATION_STRING
            == ( flags & TableModel::COLLATION_STRING ) )
    {

        for ( int i = 0; i < items_.size (); i++ )
        {
            QList<Collation> collations = items_.at ( i ).getCollations ();
            for ( int j = 0; j < collations.size (); j++ )
            {
                // If the collation number isn't in collationIndices
                // it happens nothing.
                if ( !collationIndices.isEmpty ()
                        && !collationIndices.contains ( j ) )
                    continue;

                // Replace * with .
                QString colstring =
                        collations.at ( j ).toString ().replace ( '*', '.' );
                // Insert the item at i to the list, if the collation string
                // is between begin and end.
                if ( begin.at ( 0 ) < colstring && colstring < end.at ( 0 ) )
                    list << &items_[ i ];
            }
        }
    } else
    {
        // Fills the list with all unicode letters.
        // The list size will be reduced from level to level.
        for ( int i = 0; i < items_.size (); i++ )
        {
            list << &items_[ i ];
        }

        // If the flag COLLATION_LEVEL1 is set.
        if ( ( TableModel::COLLATION_LEVEL1 & flags )
                == TableModel::COLLATION_LEVEL1 )
        {
            // If more Flags are set as begin values.
            if ( cnt >= begin.size () )
                qFatal ( "addWeightRange: Not enough ranges" );

            list = QList<Letter*> ( searchByLevel ( list, 1,
                                                    weightbegin[ cnt ],
                                                    weightend[ cnt ],
                                                    collationIndices ) );
            cnt++;
        }

        // If the flag COLLATION_LEVEL2 is set.
        if ( ( TableModel::COLLATION_LEVEL2 & flags )
                == TableModel::COLLATION_LEVEL2 )
        {
            // If more Flags are set as begin values.
            if ( cnt >= begin.size () )
                qFatal ( "addWeightRange: Not enough ranges" );
            list = QList<Letter*> ( searchByLevel ( list, 2,
                                                    weightbegin[ cnt ],
                                                    weightend[ cnt ],
                                                    collationIndices ) );
            cnt++;
        }

        // If the flag COLLATION_LEVEL3 is set.
        if ( ( TableModel::COLLATION_LEVEL3 & flags )
                == TableModel::COLLATION_LEVEL3 )
        {
            // If more Flags are set as begin values.
            if ( cnt >= begin.size () )
                qFatal ( "addWeightRange: Not enough ranges" );
            list = QList<Letter*> ( searchByLevel ( list, 3,
                                                    weightbegin[ cnt ],
                                                    weightend[ cnt ],
                                                    collationIndices ) );
            cnt++;
        }
        // If the flag COLLATION_LEVEL4 is set.
        if ( ( TableModel::COLLATION_LEVEL4 & flags )
                == TableModel::COLLATION_LEVEL4 )
        {
            // If more Flags are set as begin values.
            if ( cnt >= begin.size () )
                qFatal ( "addWeightRange: Not enough ranges" );
            list = QList<Letter*> ( searchByLevel ( list, 4,
                                                    weightbegin[ cnt ],
                                                    weightend[ cnt ],
                                                    collationIndices ) );
            cnt++;
        }
    }
    // Checks if the flags COLLATION_VARIABLE is set.
    if ( TableModel::COLLATION_VARIABLE == ( flags
            & TableModel::COLLATION_VARIABLE ) )
    {
        // Insert all letters, where the collations are variable.
        for ( int i = 0; i < list.size (); i++ )
        {
            QList<Collation> collations = list.at ( i )->getCollations ();
            for ( int j = 0; j < collations.size (); j++ )
            {

                // If the collation number isn't in collationIndices
                // it happens nothing.
                if ( !collationIndices.isEmpty ()
                        && !collationIndices.contains ( j ) )
                    continue;

                Collation col = collations.at ( j );
                if ( col.isVariable () )
                    visibleItems_ << list.at ( i );
            }
        }
    } else if ( TableModel::COLLATION_NOTVARIABLE == ( flags
            & TableModel::COLLATION_NOTVARIABLE ) )
    {
        for ( int i = 0; i < list.size (); i++ )
        {
            QList<Collation> collations = list.at ( i )->getCollations ();
            for ( int j = 0; j < collations.size (); j++ )
            {

                // If the collation number isn't in collationIndices
                // it happens nothing.
                if ( !collationIndices.isEmpty ()
                        && !collationIndices.contains ( j ) )
                    continue;

                Collation col = collations.at ( j );
                if ( !col.isVariable () )
                    visibleItems_ << list.at ( i );
            }
        }
    } else
    {
        // Insert all letters from list to the visible table.
        for ( int i = 0; i < list.size (); i++ )
        {
            visibleItems_ << list.at ( i );
        }
    }
    recalcLayout ();
    emit visibleChanged ();
}

/**
 * \brief Clears the blocks.
 *
 * Emits blocksChanged and visibleChanged.
 */
void TableModel::clearBlocks ()
{
    blocks_.clear     ();
    blockItems_.clear ();

    emit blocksChanged  ( blocks_ );
    emit visibleChanged ();
}

/**
 * \brief Clears the Model, deletes everything!
 *
 * Emits blocksChanged and visibleChanged.
 */
void TableModel::clearModel ()
{
    items_.clear      ();
    clearVisible      ();
    blocks_.clear     ();
    indexMap.clear    ();
    blockItems_.clear ();

    emit blocksChanged ( blocks_ );
    emit visibleChanged ();
}

/**
 * \brief Clears the table.
 *
 * Emits visibleChanged.
 */
void TableModel::clearTable ()
{
    items_.clear   ();
    clearVisible   ();
    indexMap.clear ();

    emit visibleChanged ();
}

/**
 * \brief Clears the view, does not delete the model items or blocks.
 *
 * Emits visibleChanged.
 */
void TableModel::clearVisible (bool clearBlockselection)
{
    if(clearBlockselection){
        blockList_->clearSelection ();
    }
    visibleItems_.clear ();
    recalcLayout ();
    oldRows_ = oldColumns_ = 0;
    emit visibleChanged ();
}

/**
 * \brief Getter for the number of columns the view should paint.
 *
 * To calculate how many columns the view can paint, it is neccessary,
 * that the view emits a resized signal with the column count.
 * (see viewResized())
 *
 * \return Number of columns.
 */
int TableModel::columnCount ( const QModelIndex & /*parent*/) const
{
    return ( viewColumns_ > visibleItems_.size () ) ? visibleItems_.size ()
                                                    : viewColumns_;
}

/**
 * \brief Checks if a specific letter is contained in the model.
 *
 * Uses the contains method of QList, which uses the operator==() method
 * of Letter. Therefore only the codepoint is compared.
 *
 * \param letter The letter to check for.
 * \return If true the model contains a Letter with the same codepoint.
 */
bool TableModel::contains ( Letter letter )
{
    return items_.contains ( letter );
}

/**
 * \brief Getter for the data of an item.
 *
 * \param index Index of the item of which data should be returned.
 * \param role Role of data to get (see enum Qt::ItemDataRole).
 * \return Data of the item, will be invalid if index or role are invalid.
 */
QVariant TableModel::data ( const QModelIndex & index, int role ) const
{
    if ( !index.isValid () || visibleItems_.isEmpty () )
        return QVariant ();

    int testIndex = index.row () * columnCount () + index.column ();
    if ( testIndex >= visibleItems_.size () )
        return QVariant ();

    if ( role == Qt::TextAlignmentRole )
    {
        return int( Qt::AlignHCenter | Qt::AlignVCenter );
    } else if ( role == Qt::DisplayRole )
    {
        return visibleItems_.at ( testIndex )->text ();
    } else if ( role == Qt::ToolTipRole )
    {
        return visibleItems_.at ( testIndex )->toolTip ();
    } else if ( role == Qt::BackgroundRole )
    {
        return visibleItems_.at ( testIndex )->background ();
    }

    return QVariant ();
}

/**
 * \brief Getter for the UCQBlockList that displays the blocks.
 *
 * \return UCQBlockList that displays the blocks.
 */
UCQBlockList* TableModel::getBlockList ()
{
    return blockList_;
}

/**
 * \brief Getter for all blocknames in the model.
 *
 * \return List of blocknames in the model.
 */
QStringList TableModel::getBlockNames ()
{
    return QStringList ( blocks_.uniqueKeys () );
}

/**
 * \brief Getter for a list of pointers on the items of the model.
 *
 * This method is supposed to be used for lookup, not for modification.
 *
 * DO NOT MODIFY THE LETTERS!
 *
 * \return list of pointer on the items in the model.
 */
QList<Letter*> TableModel::getItems ()
{
    QList<Letter*> itempointer;
    for ( int i = 0; i < items_.size (); i++ )
    {
        itempointer << &items_[ i ];
    }
    return itempointer;
}

/**
 * \brief Returns the letter of the model with the given index.
 *
 * \param index The index of the letter to get.
 * \return Letter which is at the index.
 */
Letter * TableModel::getLetter ( QModelIndex index )
{
    return getVisibleLetter ( index );
}

/**
 * \brief Return the letters with the indices from QModelIndexList.
 *
 * \param indexlist List of letters to get.
 * \return QList<Letter*> QList with all Letters with the indices.
 */
QList<Letter *> TableModel::getLetters ( QModelIndexList indexlist )
{
    QList<Letter *> list;
    for ( int i = 0; i < indexlist.size (); i++ )
    {
        list << getVisibleLetter ( indexlist.at ( i ) );
    }
    return list;
}

/**
 * \brief Getter for the pointers on the model items that are visible.
 *
 * \return Pointers on the model items that are visible.
 */
QList<Letter*> TableModel::getVisibleItems ()
{
    return visibleItems_;
}

/**
 * \brief Getter for a letter by index.
 *
 * This method is supposed to be used for lookup, not for modifcation.
 *
 * DO NOT MODIFY THE LETTER!
 *
 * \param index Index of the letter.
 * \return Pointer on the letter at the index, returns 0 if index invalid.
 */
Letter * TableModel::getVisibleLetter ( QModelIndex index )
{
    if ( index.isValid () )
    {
        int testIndex = index.row () * columnCount () + index.column ();
        if ( testIndex >= visibleItems_.size () )
            return 0;
        return visibleItems_.at ( testIndex );
    }
    return 0;
}

/**
 * \brief Checks if the model has any blocks.
 *
 * \return True, if the model has any blocks. The blocks may be emtpy.
 */
bool TableModel::hasBlocks ()
{
    return !blocks_.isEmpty ();
}

/**
 * \brief Checks if the model has items.
 *
 * \return True, if the model has items.
 */
bool TableModel::hasTable ()
{
    return !items_.isEmpty ();
}

/**
 * \brief Checks if the visible table has items.
 *
 * \return True, if the visible table is not empty.
 */
bool TableModel::hasVisibleTable ()
{
    return !visibleItems_.isEmpty ();
}

/**
 * \brief Hides a list of QModelIndexes from the view.
 *
 * Emits visibleChanged.
 *
 * \param list List of QModelIndexes to hide.
 */
void TableModel::hideIndexes ( QModelIndexList list )
{
    QList<Letter *> letters;
    bool changed = false;
    for ( int i = 0; i < list.size (); ++i )
    {
        letters.append ( getVisibleLetter ( list.at ( i ) ) );
    }

    for ( int i = 0; i < letters.size (); ++i )
    {
        if ( visibleItems_.contains ( letters.at ( i ) ) )
        {
            visibleItems_.removeAll ( letters.at ( i ) );
            changed = true;
        }
    }
    if ( changed )
    {
        recalcLayout ();
        emit visibleChanged ();
    }
}

/**
 * \brief Hide letters from the visible table which codepoints fall into the
 *        given range.
 *
 * Only the first value of the codepoint is taken into account.
 * Emits visibleChanged.
 *
 * \param start First/lowest codepoint in the range.
 * \param end Last/highest codepoint in the range.
 *
 * \see addRange()
 */
void TableModel::hideRange ( uint start, uint end )
{
    for ( int i = 0; i < visibleItems_.size (); ++i )
    {
        if ( !visibleItems_.at ( i )->getCodepoint ().isEmpty ()
                && visibleItems_.at ( i )->getCodepoint ().at ( 0 ) >= start
                && visibleItems_.at ( i )->getCodepoint ().at ( 0 ) <= end )
        {
            visibleItems_.removeAt ( i );
            --i;
        }
    }
    recalcLayout ();
    emit visibleChanged ();
}

/**
 * \brief Getter for a QModelIndex for a letter by row and column in the view.
 *
 * \param row Row in the view.
 * \param column Column in the view.
 * \return QModelIndex in the model.
 */
QModelIndex TableModel::index ( int row, int column, const QModelIndex& /*parent*/) const
{
    int testIndex = row * columnCount () + column;
    if ( testIndex >= visibleItems_.size () )
        return QModelIndex ();

    Letter *l = visibleItems_.at ( testIndex );
    return createIndex ( row, column, l );
}

/**
 * \brief Getter for an QModelIndex of a letter.
 *
 * \param letter Letter which to get the index for. The pointer must be a valid
 *               pointer on an item in the model. Use getItems() to get all
 *               available pointers.
 * \return QModelIndex of the letter in the model
 */
QModelIndex TableModel::index ( Letter * letter )
{
    if ( !visibleItems_.contains ( letter ) )
        return QModelIndex ();
    int itemIndex = visibleItems_.indexOf ( letter, 0 );
    int row = itemIndex / columnCount ();
    int col = itemIndex - columnCount () * row;
    return createIndex ( row, col, letter );
}

/**
 * \brief Recalculates the row- and columncount and emits viewResizeCells().
 *
 * This method is called when viewResized() signal was received from the view.
 * It recalculates the number of columns and rows. This is necessary, because
 * in Qt the model defines how many columns the view should have.
 * Emits layooutChanged and viewResizeCells.
 */
void TableModel::recalcLayout ()
{
    QModelIndex tempIndex = createIndex ( 0, 0, 0 );
    int cols = columnCount ();
    int rows = rowCount ();
    emit layoutAboutToBeChanged ();
    if ( rows > 0 && cols > 0 && ( cols != oldColumns_ || rows != oldRows_ ) )
    {
        if ( cols > oldColumns_ )
        {
            beginInsertColumns ( tempIndex, oldColumns_, cols - 1 );
            endInsertColumns ();
        }
        if ( cols < oldColumns_ )
        {
            beginRemoveColumns ( tempIndex, cols, oldColumns_ - 1 );
            endRemoveColumns ();
        }
        if ( rows > oldRows_ )
        {
            beginInsertRows ( tempIndex, oldRows_, rows - 1 );
            endInsertRows ();
        }
        if ( rows < oldRows_ )
        {
            beginRemoveRows ( tempIndex, rows, oldRows_ - 1 );
            endRemoveRows ();
        }
        oldColumns_ = cols;
        oldRows_ = rows;
    }
    emit layoutChanged ();
    emit viewResizeCells ();
}

/**
 * \brief Getter for the number of rows the view should paint.
 *
 * This method uses columnCount().
 *
 * \return Number of rows.
 */
int TableModel::rowCount ( const QModelIndex & /*parent*/) const
{
    int rows = 0;
    if ( !visibleItems_.isEmpty () )
    {
        if ( columnCount () > 0 )
            rows = visibleItems_.size () / columnCount ();
        int c = columnCount ();
        while ( rows * c < visibleItems_.size () )
            rows++;
    }
    return rows;
}

/**
 * \brief Method that performs the search for addWeightRange.
 *
 * This is only a helper method for addWeightRange
 * See addWeightRange for details.
 *
 * \param list List of Letter to search in.
 * \param level Level of the collations to search.
 * \param begin First/lowest collation.
 * \param end Last/highest collation.
 * \param collationIndices Defines which collations define the range it letters
 *                         have multiple collations.
 * \return A list with hits of the search.
 * \see addWeightRange()
 */
QList<Letter*> TableModel::searchByLevel ( QList<Letter*> & list, int level,
                                           uint begin, uint end,
                                           QList<int> collationIndices )
{

    QList<Letter*> newList;
    for ( int i = 0; i < list.size (); i++ )
    {
        QList<Collation> collations = list.at ( i )->getCollations ();

        for ( int j = 0; j < collations.size (); j++ )
        {
            // If the collation number isn't in collationIndices
            // it happens nothing.
            if ( !collationIndices.isEmpty ()
                    && !collationIndices.contains ( j ) )
                continue;

            Collation collation = collations.at ( j );
            QList<uint> weights = collation.getWeights ();
            // Checks if the weights are between begin and end.
            if ( begin <= weights.at ( level - 1 ) && weights.at ( level - 1 )
                    <= end )
            {
                newList << list.at ( i );
                break;
            }
        }
    }

    return newList;
}

/**
 * \brief Setter for the UCQBlockList that displays the blocks.
 *
 * \param list QListWidget that displays the blocks
 */
void TableModel::setBlockList ( UCQBlockList* list )
{
    blockList_ = list;
}

/**
 * \brief Shows complete content in the view.
 *
 * This method clears the visibleItems list and populates it with pointers
 * to all items in the model.
 * Emits visibleChanged.
 */
void TableModel::showAll ()
{
    visibleItems_.clear ();
    for ( int i = 0; i < items_.size (); ++i )
    {
        visibleItems_.append ( &items_[ i ] );
    }
    recalcLayout ();
    emit visibleChanged ();
}

/**
 * \brief Shows the letters of the given blocks in the view.
 *
 * Emits visibleChanged.
 *
 * \param blockList Names of the blocks to show. If the list contains "Show all", all
 *                  available blocks will be shown. If the blockList is empty
 *                  the view will be cleared.
 *                  (If possible use showAll() and clearVisible() in those
 *                   cases.)
 */
void TableModel::showBlocks ( QStringList blockList )
{
    visibleItems_.clear ();
    if ( !blockList.isEmpty () )
    {
        if ( blockList.contains ( tr ( "Show all" ) ) )
        {
            showAll ();
        } else
        {
            QList<Letter*> block;
            for ( int i = 0; i < blockList.size (); ++i )
            {
                block = blockItems_.values ( blockList.at ( i ) );
                if ( !block.isEmpty () )
                    visibleItems_.append ( block );
            }
        }
    } else
    {
        oldRows_ = oldColumns_ = 0;
    }
    recalcLayout ();
    emit visibleChanged ();
}

/**
 * \brief Reduces the contents of the view on the indices in the list.
 *
 * Emits visibleChanged.
 *
 * \param list List of QModelIndexes to keep.
 */
void TableModel::showIndexes ( QModelIndexList list )
{
    QList<Letter *> letters;
    // check all letters, but do not repaint if they are all already visible
    bool changed = false;
    for ( int i = 0; i < list.size (); ++i )
    {
        letters.append ( getVisibleLetter ( list.at ( i ) ) );
    }
    if ( !letters.isEmpty () )
    {
        visibleItems_.clear ();
        for ( int i = 0; i < letters.size (); ++i )
        {
            if ( !visibleItems_.contains ( letters.at ( i ) ) )
            {
                visibleItems_.append ( letters.at ( i ) );
                changed = true;
            }
        }
        if ( changed )
        {
            recalcLayout ();
            emit visibleChanged ();
        }
    }
}

/**
 * \brief Checks if the blocks contain all items they should.
 *
 * This funtion is called when a new block is added to the model to populate it
 * with the right letters. The method is not needed when a new letter is added,
 * because it will be added to right block.
 */
void TableModel::updateBlocks ()
{
    if ( !blocks_.isEmpty () )
    {
        QStringList blockNames = blocks_.uniqueKeys ();
        QStringList blockKeys = blockItems_.uniqueKeys ();
        uint code;
        QString blockName;
        Letter * l;
        QList<uint> startEnd;
        blockItems_.clear ();
        for ( int j = 0; j < blockNames.size (); ++j )
        {
            // get the name and the boundaries for the block
            blockName = blockNames.at ( j );
            startEnd = blocks_.values ( blockName );
            if ( startEnd.size () < 2 )
                continue;

            // Find all letters that belong into the block
            // insertMulti puts new items in front
            for ( int i = items_.size () - 1; i >= 0; --i )
            {
                l = &items_[ i ];
                if ( l->getCodepoint ().isEmpty () )
                    continue;

                code = l->getCodepoint ().at ( 0 );

                if ( code >= startEnd.at ( 0 ) && code <= startEnd.at ( 1 ) )
                {
                    blockItems_.insertMulti ( blockName, l );
                    items_[ i ].setBlock ( blockName );
                }
            }
        }
    }
}

/**
 * \brief Slot for resize events of the view.
 *
 * Calls recalcLayout() which applies the new column count.
 *
 * \param columnCount The new column count.
 */
void TableModel::viewResized ( int columnCount )
{
    viewColumns_ = columnCount;
    recalcLayout ();
}

///**
// * \brief Shows only the letters included in the given font in the visible
// *        table.
// *    DOESNT WORK SEE Letter::inFont()...
// * \param font The font of which the letters should be shown.
// */
//void TableModel::showLettersOfFont(const QFont& font){
//    visibleItems_.clear();
//    for(int i = 0; i < items_.size(); ++i){
//        if(items_[i].inFont(font)){
//            visibleItems_.append(&items_[i]);
//        }
//    }
//    recalcLayout ();
//    emit visibleChanged ();
//}
